#pragma once

#define METEOSAT_8_SCID 321
#define METEOSAT_9_SCID 322
#define METEOSAT_10_SCID 323
#define METEOSAT_11_SCID 324

#define METEOSAT_8_NORAD 27509
#define METEOSAT_9_NORAD 28912
#define METEOSAT_10_NORAD 38552
#define METEOSAT_11_NORAD 40732
