/*
 * Generated by asn1c-0.9.28 (http://lionet.info/asn1c)
 * From ASN.1 module "FANSACTwoWayDataLinkCommunications"
 * 	found in "../../../libacars.asn1/fans-cpdlc.asn1"
 * 	`asn1c -fcompound-names -fincludes-quoted -gen-PER`
 */

#ifndef	_FANSDistanceOffset_H_
#define	_FANSDistanceOffset_H_


#include "asn_application.h"

/* Including external dependencies */
#include "FANSDistanceOffsetNm.h"
#include "FANSDistanceOffsetKm.h"
#include "constr_CHOICE.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Dependencies */
typedef enum FANSDistanceOffset_PR {
	FANSDistanceOffset_PR_NOTHING,	/* No components present */
	FANSDistanceOffset_PR_distanceOffsetNm,
	FANSDistanceOffset_PR_distanceOffsetKm
} FANSDistanceOffset_PR;

/* FANSDistanceOffset */
typedef struct FANSDistanceOffset {
	FANSDistanceOffset_PR present;
	union FANSDistanceOffset_u {
		FANSDistanceOffsetNm_t	 distanceOffsetNm;
		FANSDistanceOffsetKm_t	 distanceOffsetKm;
	} choice;
	
	/* Context for parsing across buffer boundaries */
	asn_struct_ctx_t _asn_ctx;
} FANSDistanceOffset_t;

/* Implementation */
extern asn_TYPE_descriptor_t asn_DEF_FANSDistanceOffset;

#ifdef __cplusplus
}
#endif

#endif	/* _FANSDistanceOffset_H_ */
#include "asn_internal.h"
